
/*
* dataUtility.cpp contains the implementations of the functions declared in Data_Utility.h.
*/

#include "dataUtility.h"


#ifndef WIN32

/*
* GetCurTime is used to get the current running time in the current process.
*
* @Param curTime gets back the time information.
*
* @Return void.
*/
void get_cur_time( struct rusage* curTime)
{
	int ret = getrusage( RUSAGE_SELF, curTime);
	if( ret != 0)
	{
		fprintf( stderr, "The running time info couldn't be collected successfully.\n");
		//data_release( 2);
		exit( 0);
	}
}

/*
* GetTime is used to get the 'float' format time from the start and end rusage structure.
* 
* @Param timeStart, timeEnd indicate the two time points.
* @Param userTime, sysTime get back the time information.
* 
* @Return void.
*/
void collect_time( struct rusage* timeStart, struct rusage* timeEnd, float* userTime, float* sysTime)
{
	(*userTime) = ((float)(timeEnd->ru_utime.tv_sec - timeStart->ru_utime.tv_sec)) + 
		((float)(timeEnd->ru_utime.tv_usec - timeStart->ru_utime.tv_usec)) * 1e-6;
	(*sysTime) = ((float)(timeEnd->ru_stime.tv_sec - timeStart->ru_stime.tv_sec)) +
		((float)(timeEnd->ru_stime.tv_usec - timeStart->ru_stime.tv_usec)) * 1e-6;
}

#endif



/*
 *	Generates a number between [min, max] randomly.
 */
int rand_i( int min, int max)
{
	int i_rand;
	double ratio;

	if( min > max)
		return 0;

	if( max - min > RAND_MAX)
	{
		//printf( "rand_i: [%i, %i] out of range.\n", min, max);
		//exit( 0);

		ratio = rand_f( 0.0, 1.0);
		i_rand = min + ( int)( ratio * ( max - min));

		return i_rand;
	}

	//srand( time( NULL));

	if( min == max)
		i_rand = 0;
	else
		i_rand = rand( ) % ( max - min + 1);

	return i_rand + min;
}

/*
 *	Generate a random float value in [min_v, max_v].
 */
double rand_f( double min_v, double max_v)
{
	double rand_v;

	if( min_v > max_v)
		return 0;

	//srand( time(NULL));

	rand_v = double( rand( )) / (double)RAND_MAX; //rand_v falls in [0,1] randomly.

	return ( max_v - min_v) * rand_v + min_v;
}

/*
 *	Check whether an variable has been generated before.
 *
 */
bool is_old( float* rand_v, int cur, float var)
{
	int i;
	
	for( i=0; i<cur; i++)
	{
		if( rand_v[ i] == var)
			return true;
	}

	return false;
}

/*
* dReadConfig reads the configuration information for Data_Utility functions.
*
* @Return the configuration info with d_config_t structure.
*/
d_config_t* dReadConfig( )
{
	char des[MAX_DESCRIPTION_LENG];
	char tag[4];
	d_config_t* d_cfg = (d_config_t*)malloc( sizeof( d_config_t));

	FILE* c_fp = fopen( D_CONFIG_FILE, "r");
	if( c_fp == NULL)
	{
		fprintf( stderr, "The d_configuration file cannot be open.\n");
		exit(0);
	}

	//Reading the information.
	fscanf( c_fp, "%s%s", des, d_cfg->p_obj);
	fscanf( c_fp, "%s%s", des, d_cfg->o_obj);
	fscanf( c_fp, "%s%s", des, tag);

	if( strcmp( tag, "yes") != 0)
	{
		fscanf( c_fp, "%s%s", des, d_cfg->p_weight);
		fscanf( c_fp, "%s%s", des, d_cfg->o_weight);
		d_cfg->w_combined = 0;
	}
	else
		d_cfg->w_combined = 1;

	fscanf( c_fp, "%s%i", des, &d_cfg->dim);
	fscanf( c_fp, "%s%s", des, d_cfg->matching);
	fscanf( c_fp, "%s%s", des, d_cfg->stat_file);
/*	fscanf( c_fp, "%s%i", des, d_cfg->n_p_obj);
	fscanf( c_fp, "%s%i", des, d_cfg->n_o_obj);*/

	//Releasing the related the resource.
	fclose( c_fp);

	return d_cfg;
}

/*
* dReadData reads the raw data information for P and O objects according to the configuration info.
*
* @d_cfg indicates the configuration information of the input data.
*
* @Return the data info with data_t structure.
*/
data_t* dReadData( d_config_t* d_cfg)
{
	int i, j;
	data_t* data_v = (data_t*)malloc( sizeof( data_t));

	FILE *p_fp, *o_fp, *p_w_fp, *o_w_fp;
	p_fp = fopen( d_cfg->p_obj, "r");
	o_fp = fopen( d_cfg->o_obj, "r");
	if( p_fp == NULL || o_fp == NULL)
	{
		fprintf( stderr, "The p_obj/o_obj file cannot be open.\n");
		exit(0);
	}

	if( d_cfg->w_combined == 0)		//For the un-combined case.
	{
		p_w_fp = fopen( d_cfg->p_weight, "r");
		o_w_fp = fopen( d_cfg->o_weight, "r");
		if( p_w_fp == NULL || o_w_fp == NULL)
		{
			fprintf( stderr, "The p_weigh/o_weight file cannot be open.\n");
			exit(0);
		}
	}

	//Assigning partial info of data_v.
	data_v->dim = d_cfg->dim;
/*	data_v->n_p_objs = d_cfg->n_p_obj;
	data_v->n_o_objs = d_cfg->n_o_obj;*/
	fscanf( p_fp, "%i", &(data_v->n_p_objs));
	fscanf( o_fp, "%i", &(data_v->n_o_objs));
	
	data_v->p_objs = (obj_t*)malloc( sizeof(obj_t) * data_v->n_p_objs);
	data_v->o_objs = (obj_t*)malloc( sizeof(obj_t) * data_v->n_o_objs);
	
	//Reading P_objects.
	for( i=0; i<data_v->n_p_objs; i++)
	{
		data_v->p_objs[i].axis = (FLOAT*)malloc( sizeof(FLOAT) * data_v->dim);
		for( j=0; j<data_v->dim; j++)
		{
			fscanf( p_fp, "%f", data_v->p_objs[i].axis + j);
		}
		if( d_cfg->w_combined == 1)
			fscanf( p_fp, "%f", &(data_v->p_objs[i].weight));
		else
			fscanf( p_w_fp, "%f", &(data_v->p_objs[i].weight));
	}//
	
	//Reading O_objects.
	for( i=0; i<data_v->n_o_objs; i++)
	{
		data_v->o_objs[i].axis = (FLOAT*)malloc( sizeof(FLOAT) * data_v->dim);
		for( j=0; j<data_v->dim; j++)
		{
			fscanf( o_fp, "%f", data_v->o_objs[i].axis + j);
		}
		if( d_cfg->w_combined == 1)
			fscanf( o_fp, "%f", &(data_v->o_objs[i].weight));
		else
			fscanf( o_w_fp, "%f", &(data_v->o_objs[i].weight));
	}

	//Releasing the related resources.
	fclose( p_fp);
	fclose( o_fp);
	if( d_cfg->w_combined == 0)
	{
		fclose( p_w_fp);
		fclose( o_w_fp);
	}

	return data_v;
}


/*
* data_release releases the data_t structure indicated by "data_v".
*
* @data_v indicates the data_t structure.
*/
void data_release( data_t* data_v)
{
	int i;

	//Releasing the P objects.
	for( i=0; i<data_v->n_p_objs; i++)
	{
		free( data_v->p_objs[i].axis);
	}

	free( data_v->p_objs);

	//Releasing the O objects.
	for( i=0; i<data_v->n_o_objs; i++)
	{
		free( data_v->o_objs[i].axis);
	}

	free( data_v->o_objs);
}


//-----------------------------

_config_t* emp_read_config( )
{
	//char des[MAX_DESCRIPTION_LENG];
	//char tag[4];
	_config_t* cfg = (_config_t*)malloc( sizeof( _config_t));

	FILE* c_fp = fopen( EMP_CONFIG_FILE, "r");
	if( c_fp == NULL)
	{
		fprintf( stderr, "The config.txt file cannot be open.\n");
		exit(0);
	}

	//Reading the information.
	fscanf( c_fp, "%s%s", cfg->o_file, cfg->p_file);
	fscanf( c_fp, "%i%i", &cfg->n_o, &cfg->n_p);
	fscanf( c_fp, "%i", &cfg->dim);
	
	//fscanf( c_fp, "%f", &cfg->n_r);
	//fscanf( c_fp, "%f", &cfg->w_r);

	/*
	fscanf( c_fp, "%i%i%i%i", &cfg->thr_opt, 
								&cfg->speed_opt, 
								&cfg->swap_opt, 
								&cfg->graph_tag);
	*/
	fscanf( c_fp, "%i%i", &cfg->thr_opt, 
								&cfg->swap_opt);

	cfg->speed_opt = 2;
	cfg->graph_tag = 0;
	
	
	//Releasing the related the resource.
	fclose( c_fp);

	return cfg;
}

/*
 *
 */
data_t* emp_read_data( _config_t* cfg)
{
	int i, j;
	char  des[10];
	FLOAT o_w_sum = 0, p_w_sum = 0;

	FILE *p_fp, *o_fp;

	data_t* data_v = (data_t*)malloc( sizeof( data_t));
	memset( data_v, 0, sizeof( data_t));

	p_fp = fopen( cfg->p_file, "r");
	o_fp = fopen( cfg->o_file, "r");
	if( p_fp == NULL || o_fp == NULL)
	{
		fprintf( stderr, "The p_obj/o_obj file cannot be open.\n");
		exit(0);
	}

	//Assigning partial info of data_v.
	data_v->dim = cfg->dim;
	data_v->n_p_objs = cfg->n_p;
	data_v->n_o_objs = cfg->n_o;
	
	data_v->p_objs = (obj_t*)malloc( sizeof(obj_t) * data_v->n_p_objs);
	memset( data_v->p_objs, 0, sizeof( obj_t) * data_v->n_p_objs);

	data_v->o_objs = (obj_t*)malloc( sizeof(obj_t) * data_v->n_o_objs);
	memset( data_v->o_objs, 0, sizeof( obj_t) * data_v->n_o_objs);
	
	//Reading P_objects.
	for( i=0; i<data_v->n_p_objs; i++)
	{
		data_v->p_objs[i].axis = ( FLOAT*)malloc( sizeof(FLOAT) * data_v->dim);
		memset( data_v->p_objs[ i].axis, 0, sizeof( FLOAT) * data_v->dim);

		data_v->p_objs[ i].MBR = ( range_t*)malloc( sizeof( range_t) * data_v->dim);
		memset( data_v->p_objs[ i].MBR, 0, sizeof( range_t) * data_v->dim);

		fscanf( p_fp, "%s%f", des, &( data_v->p_objs[ i].weight));

		//if( DISCRITITIZE_TAG)
		//	data_v->p_objs[ i].weight -= ( (int)data_v->p_objs[ i].weight) % DISCRITITIZE_SCALE;
		
		//s
		if( data_v->p_objs[ i].weight == 0)
			data_v->p_objs[ i].weight = 1;
		//s

		data_v->p_objs[ i].id = i;
/*t*
		data_v->p_objs[i].weight = cfg->n_r * cfg->w_r;
/*t*/
		p_w_sum += data_v->p_objs[i].weight;
		for( j=0; j<data_v->dim; j++)
		{
			fscanf( p_fp, "%f", data_v->p_objs[i].axis + j);
			
			if( PERTUBATION_TAG == 1)
				data_v->p_objs[ i].axis[ j] += (float)rand_f( PERTUBATION_INTERVAL_LEFT, PERTUBATION_INTERVAL_RIGHT);
		
			data_v->p_objs[ i].MBR[ j].min = data_v->p_objs[ i].axis[ j];
			data_v->p_objs[ i].MBR[ j].max = data_v->p_objs[ i].MBR[ j].min;
		}
	}//
	
	//Reading O_objects.
	for( i=0; i<data_v->n_o_objs; i++)
	{
		data_v->o_objs[i].axis = (FLOAT*)malloc( sizeof(FLOAT) * data_v->dim);
		memset( data_v->o_objs[ i].axis, 0, sizeof( FLOAT) * data_v->dim);

		fscanf( o_fp, "%s%f", des, &(data_v->o_objs[i].weight));

		if( DISCRITITIZE_TAG)
			data_v->o_objs[ i].weight -= ( (int)data_v->o_objs[ i].weight) % DISCRITITIZE_SCALE;
		
		//s
		if( data_v->o_objs[ i].weight == 0)
			data_v->o_objs[ i].weight = 1;
		//s

		data_v->o_objs[ i].id = i;
/*t*
		data_v->o_objs[i].weight = 1;
/*t*/
		o_w_sum += data_v->o_objs[i].weight;
		for( j=0; j<data_v->dim; j++)
		{
			fscanf( o_fp, "%f", data_v->o_objs[i].axis + j);

			if( PERTUBATION_TAG == 1)
				data_v->o_objs[ i].axis[ j] += ( float)rand_f( PERTUBATION_INTERVAL_LEFT, PERTUBATION_INTERVAL_RIGHT);
		}
	}

	//Testing the assumption condition.
	if( p_w_sum < o_w_sum)
	{
		fprintf( stderr, "The assumption that p_w_sum is at leat o_w_sum is not satisfied.\n");
		exit(0);
	}


	cfg->n_r = ( float)cfg->n_o / cfg->n_p;
	cfg->w_r = ( float)p_w_sum / o_w_sum;

	//Releasing the related resources.
	fclose( p_fp);
	fclose( o_fp);

	return data_v;
}

/*
 *	Print the statistics information (the stat_t structure).
 */
void print_stat( DOUBLE mmd, stat_t stat_v)
{
	FILE* s_fp;

	if( (s_fp = fopen( STATISTIC_FILE, "w")) == NULL)
	{
		fprintf( stderr, "The statistic file cannot be open.\n");
		exit(0);
	}

	//mmd.
	fprintf( s_fp, "%lf\n\n", mmd); 

	//time.
	//fprintf( s_fp, "%f\n%f\n%f\n\n\n", stat_v.preprocess_time, stat_v.matching_time, stat_v.total_time);
	fprintf( s_fp, "%f\n\n", stat_v.total_time);
	
	//memory.
	fprintf( s_fp, "%-.0f\n\n\n", stat_v.memory_max / MEMORY_SCALE);

/*
	//Part I: Threshold_Adapt.
	fprintf( s_fp, "%f\n%f\n\n\n", stat_v.n_path, stat_v.path_length_aver);	
	fprintf( s_fp, "%f\n%f\n\n\n", stat_v.que_length_max, stat_v.que_length_aver);

	//Part II: Swap_Chain.
	//n_redcution & n_swap.
	fprintf( s_fp, "%-.0f\n\n\n%-.0f\n%-.0f\n%-.0f\n\n\n", stat_v.n_reduction, stat_v.n_swap.aver, 
												stat_v.n_swap.min, stat_v.n_swap.max);
	
	//n_swap_que.
	fprintf( s_fp, "%-.0f\n%-.0f\n%-.0f\n\n\n", stat_v.n_swap_que.aver, stat_v.n_swap_que.min, stat_v.n_swap_que.max);

	//n_expension.
	fprintf( s_fp, "%-.0f\n%-.0f\n%-.0f\n\n\n", stat_v.n_expension.aver, stat_v.n_expension.min, stat_v.n_expension.max);

	//n_edge.
	fprintf( s_fp, "%-.0f\n%-.0f\n\n\n", stat_v.n_ini_edge, stat_v.n_edge_aver);

	//n_pair.
	fprintf( s_fp, "%-.0f\n%-.0f\n\n\n", stat_v.n_ini_pair, stat_v.n_pair_max);

	//R_tree_build_time.
	fprintf( s_fp, "%f\n", stat_v.R_tree_build_time);

*/
	fclose( s_fp);
}
