/*
  This code was written by Raymond Chi-Wing Wong.
  Please read the following.

    Raymond Chi-Wing Wong, Yufei Tao, Ada Wai-Chee Fu and Xiaokui Xiao. 
    "On Efficient Spatial Matching", 
    the 33rd International Conference on Very Large Data Bases (VLDB'07), 
    Vienna, Austria on 23-28 Sept, 2007 

*/

#include "statistics.h"

statistics *statistics_new()
{
	statistics *returnValue;

	returnValue = (statistics *) malloc(sizeof(statistics));
	memset(returnValue, 0, sizeof(statistics));

	returnValue->chainLen_min = 1000000;
	returnValue->noOfCoupleForO_min = 1000000;
	returnValue->noOfCoupleForP_max = 1000000;

	return returnValue;
}

void statistics_free(statistics **aStat)
{
	free((*aStat));
}

void statistics_update_chainLen(statistics *aStat, int chainLen)
{
	if (aStat->chainLen_max < chainLen)
	{
		aStat->chainLen_max = chainLen;
	}

	if (aStat->chainLen_min > chainLen)
	{
		aStat->chainLen_min = chainLen;
	}

	aStat->chainLen_totalNoOfChain++;
	aStat->chainLen_totalLen += chainLen;
}

void statistics_update_noOfCoupleForO(statistics *aStat, int noOfCoupleForO)
{
	if (aStat->noOfCoupleForO_max < noOfCoupleForO)
	{
		aStat->noOfCoupleForO_max = noOfCoupleForO;
	}

	if (aStat->noOfCoupleForO_min > noOfCoupleForO)
	{
		aStat->noOfCoupleForO_min = noOfCoupleForO;
	}

	aStat->noOfCoupleForO_total += noOfCoupleForO;
}

void statistics_update_noOfCoupleForP(statistics *aStat, int noOfCoupleForP)
{
	if (aStat->noOfCoupleForP_max < noOfCoupleForP)
	{
		aStat->noOfCoupleForP_max = noOfCoupleForP;
	}

	if (aStat->noOfCoupleForP_min > noOfCoupleForP)
	{
		aStat->noOfCoupleForP_min = noOfCoupleForP;
	}

	aStat->noOfCoupleForP_total += noOfCoupleForP;
}

void statistics_finalize(int noOfPointO, int noOfPointP)
{
	globalStat->totalNoOfNNQuery = globalStat->totalNoOfNNQueryFromP + globalStat->totalNoOfNNQueryFromO;
	globalStat->totalNoOfDeletion = globalStat->totalNoOfDeletionFromP + globalStat->totalNoOfDeletionFromO;

	globalStat->ratioPerOSize = globalStat->totalNoOfNNQuery/((double) noOfPointO);
	globalStat->ratioPerPPlusOSize = globalStat->totalNoOfNNQuery/((double) noOfPointP+noOfPointO);
}

void statistics_noOfCouple(int noOfDim, twoData *aTwoData, assignment *finalAssignment, assignmentSet *assignmentSetForO, assignmentSet *assignmentSetForP)
{
	int i;
	int noOfO, noOfP;

	int noOfCouple;

	assignment **assignO;
	assignment **assignP;

	assignment *singleAssignO;
	assignment *singleAssignP;

	int weight;

	double cost;

	triple *aTriple;

	triple **aAssign;

	pointList *O;
	pointList *P;

	int indexP, indexO;

	O = aTwoData->O;
	P = aTwoData->P;

	noOfO = O->noOfPoint;
	noOfP = P->noOfPoint;
	

	// totalNoOfCouple
	globalStat->totalNoOfCouple = finalAssignment->noOfMatch;

	// totalMatchingCost and totalMatchingWeight
	aAssign = finalAssignment->list;
	for (i = 0; i < finalAssignment->noOfMatch; i++)
	{
		aTriple = aAssign[i];

		weight = aTriple->weight;

		indexO = aTriple->oid;
		indexP = aTriple->pid - noOfO;
		cost = (double) stable_distance(indexO, indexP, O, P, noOfDim);

		globalStat->totalMatchingWeight += weight;

		globalStat->totalMatchingCost += cost;
	}

	//  ******
	noOfO = assignmentSetForO->noOfAssign;
	assignO = assignmentSetForO->list;

	noOfP = assignmentSetForP->noOfAssign;
	assignP = assignmentSetForP->list;

	// noOfCoupleForO
	for (i = 0; i < noOfO; i++)
	{
		singleAssignO = assignO[i];

		noOfCouple = singleAssignO->noOfMatch;

		statistics_update_noOfCoupleForO(globalStat, noOfCouple);
	}

	// noOfCoupleForP
	for (i = 0; i < noOfP; i++)
	{
		singleAssignP = assignP[i];

		noOfCouple = singleAssignP->noOfMatch;

		statistics_update_noOfCoupleForP(globalStat, noOfCouple);
	}
}

void statistics_prefListRankList(twoData *aTwoData)
{
	pointList *O;
	pointList *P;

	int noOfPointO;
	int noOfPointP;

	O = aTwoData->O;
	P = aTwoData->P;

	noOfPointO = O->noOfPoint;
	noOfPointP = P->noOfPoint;

	globalStat->prefList_storage = noOfPointO*noOfPointP*4;
	globalStat->rankList_storage = noOfPointO*noOfPointP*4;
}

void statistics_totalStorage(configure *aConfigure)
{
	int ALG;

	int noOfPointO, noOfPointP;
	pointList *O;
	pointList *P;
	
	ALG = aConfigure->ALG;

	if (ALG == 1)
	{
		globalStat->totalStorage = globalStat->rtreeO_storage + globalStat->rtreeP_storage + globalStat->chainLen_max*4;
	}
	else if (ALG == 2)
	{
		globalStat->totalStorage = globalStat->prefList_storage + globalStat->rankList_storage;
	}
	else if (ALG == 3)
	{
		globalStat->totalStorage = globalStat->rtreeO_storage + globalStat->rtreeP_storage; // + globalStat->chainLen_max*4;
	}
	else if (ALG == 4)
	{
		O = aConfigure->O;
		P = aConfigure->P;
		noOfPointO = O->noOfPoint;
		noOfPointP = P->noOfPoint;
		globalStat->totalStorage = noOfPointP*noOfPointO*4;
	}
}

void statistics_rtreeNoOfNodeStorage(node_type *rootO, node_type *rootP, rtree_info *aInfo)
{
	int noOfNode;

	int nodeSize;

	nodeSize = rtree_nodeSize(aInfo);

	noOfNode = 0;
	noOfNodeRtree(rootO, &noOfNode, aInfo);
	globalStat->noOfNodeRtreeO = noOfNode;
	globalStat->rtreeO_storage = noOfNode*nodeSize;

	noOfNode = 0;
	noOfNodeRtree(rootP, &noOfNode, aInfo);
	globalStat->noOfNodeRtreeP = noOfNode;
	globalStat->rtreeP_storage = noOfNode*nodeSize;

}

void statistics_endMain(configure *aConfigure, twoData *aTwoData, assignment *finalAssignment, assignmentSet *assignmentSetForO, assignmentSet *assignmentSetForP)
{
	int noOfDim;
	rtree_info *rtreeInfo;

	int noOfPointO, noOfPointP;

	rtreeInfo = aConfigure->rtreeInfo;

	noOfPointO = aTwoData->O->noOfPoint;
	noOfPointP = aTwoData->P->noOfPoint;

	noOfDim = rtreeInfo->dim;
	statistics_noOfCouple(noOfDim, aTwoData, finalAssignment, assignmentSetForO, assignmentSetForP);
	statistics_prefListRankList(aTwoData);

	statistics_totalStorage(aConfigure);

	statistics_finalize(noOfPointO, noOfPointP);
}


void statistics_print(FILE *filePtr)
{
	fprintf(filePtr, "// chainLen\n");
	fprintf(filePtr, "%d\n", globalStat->chainLen_min);
	fprintf(filePtr, "%d\n", globalStat->chainLen_max);
	fprintf(filePtr, "%.10f\n", globalStat->chainLen_totalLen);
	fprintf(filePtr, "%d\n", globalStat->chainLen_totalNoOfChain);
	fprintf(filePtr, "// noOfNNQuery and noOfDeletion\n");
	fprintf(filePtr, "%d\n", globalStat->totalNoOfNNQueryFromP);
	fprintf(filePtr, "%d\n", globalStat->totalNoOfNNQueryFromO);
	fprintf(filePtr, "%d\n", globalStat->totalNoOfNNQuery);
	fprintf(filePtr, "%.10f\n", globalStat->ratioPerOSize);
	fprintf(filePtr, "%.10f\n", globalStat->ratioPerPPlusOSize);
	fprintf(filePtr, "%d\n", globalStat->totalNoOfDeletionFromP);
	fprintf(filePtr, "%d\n", globalStat->totalNoOfDeletionFromO);
	fprintf(filePtr, "%d\n", globalStat->totalNoOfDeletion);
	fprintf(filePtr, "// noOfCouple\n");
	fprintf(filePtr, "%d\n", globalStat->totalNoOfCouple);
	fprintf(filePtr, "%d\n", globalStat->noOfCoupleForO_min);
	fprintf(filePtr, "%d\n", globalStat->noOfCoupleForO_max);
	fprintf(filePtr, "%d\n", globalStat->noOfCoupleForO_total);
	fprintf(filePtr, "%d\n", globalStat->noOfCoupleForP_min);
	fprintf(filePtr, "%d\n", globalStat->noOfCoupleForP_max);
	fprintf(filePtr, "%d\n", globalStat->noOfCoupleForP_total);
	fprintf(filePtr, "// Matching cost and matching weight\n");
	fprintf(filePtr, "%.10f\n", globalStat->totalMatchingCost);
	fprintf(filePtr, "%.10f\n", globalStat->totalMatchingWeight);
	fprintf(filePtr, "// Rtree\n");
	fprintf(filePtr, "%d\n", globalStat->noOfNodeRtreeO);
	fprintf(filePtr, "%d\n", globalStat->rtreeO_storage);
	fprintf(filePtr, "%d\n", globalStat->noOfNodeRtreeP);
	fprintf(filePtr, "%d\n", globalStat->rtreeP_storage);
	fprintf(filePtr, "// prefList and rankList\n");
	fprintf(filePtr, "%d\n", globalStat->prefList_storage);
	fprintf(filePtr, "%d\n", globalStat->rankList_storage);
	fprintf(filePtr, "// totalStorage\n");
	fprintf(filePtr, "%d\n", globalStat->totalStorage);
}

